import unittest
import itertools

from myhdl import *
from myhdl_lib.fifo_speculative import fifo_speculative
from myhdl_lib import sfifo_beh
import myhdl_lib.simulation as sim


class TestSFifo(unittest.TestCase):

    COMMIT = 0
    DISCARD = 1
    DISCARD_COMMIT = 2

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def setUp(self):
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 128
        DEPTH_MAX = 101

        self.sfifo_model = None

        self.full = Signal(bool(0))
        self.we = Signal(bool(0))
        self.din = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))
        self.empty = Signal(bool(0))
        self.re = Signal(bool(0))
        self.dout = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))
        self.wr_commit = Signal(bool(0))
        self.wr_discard = Signal(bool(0))
        self.rd_commit = Signal(bool(0))
        self.rd_discard = Signal(bool(0))
        self.afull = Signal(bool(0))
        self.aempty = Signal(bool(0))
        self.afull_th = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.aempty_th = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.count = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.count_max = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.ovf = Signal(bool(0))
        self.udf = Signal(bool(0))

        self.clk = sim.Clock(val=0, period=10, units="ns")
        self.rst = sim.ResetSync(clk=self.clk, val=0, active=1)
        self.clkgen = self.clk.gen()


    def tearDown(self):
        pass

    def reset(self):
        yield self.rst.pulse(5)

    def fifo_command(self, wcmd=None, rcmd=None):
        self.sfifo_model.command(wcmd, rcmd)
        self.wr_commit.next = 1 if (wcmd == self.COMMIT  or wcmd == self.DISCARD_COMMIT) else 0
        self.wr_discard.next = 1 if (wcmd == self.DISCARD or wcmd == self.DISCARD_COMMIT) else 0
        self.rd_commit.next = 1 if (rcmd == self.COMMIT  or rcmd == self.DISCARD_COMMIT) else 0
        self.rd_discard.next = 1 if (rcmd == self.DISCARD or rcmd == self.DISCARD_COMMIT) else 0
        yield self.clk.posedge
        self.wr_commit.next = 0
        self.wr_discard.next = 0
        self.rd_commit.next = 0
        self.rd_discard.next = 0

    def fifo_write(self, val, wcmd=None, rcmd=None):
        self.sfifo_model.write(val, wcmd, rcmd)
        self.we.next = 1
        self.din.next = val
        self.wr_commit.next = 1 if (wcmd == self.COMMIT  or wcmd == self.DISCARD_COMMIT) else 0
        self.wr_discard.next = 1 if (wcmd == self.DISCARD or wcmd == self.DISCARD_COMMIT) else 0
        self.rd_commit.next = 1 if (rcmd == self.COMMIT  or rcmd == self.DISCARD_COMMIT) else 0
        self.rd_discard.next = 1 if (rcmd == self.DISCARD or rcmd == self.DISCARD_COMMIT) else 0
        yield self.clk.posedge
        self.we.next = 0
        self.din.next = 0
        self.wr_commit.next = 0
        self.wr_discard.next = 0
        self.rd_commit.next = 0
        self.rd_discard.next = 0

    def fifo_read(self, wcmd=None, rcmd=None):
        self.sfifo_model.read(wcmd, rcmd)
        self.re.next = 1
        self.wr_commit.next = 1 if (wcmd == self.COMMIT  or wcmd == self.DISCARD_COMMIT) else 0
        self.wr_discard.next = 1 if (wcmd == self.DISCARD or wcmd == self.DISCARD_COMMIT) else 0
        self.rd_commit.next = 1 if (rcmd == self.COMMIT  or rcmd == self.DISCARD_COMMIT) else 0
        self.rd_discard.next = 1 if (rcmd == self.DISCARD or rcmd == self.DISCARD_COMMIT) else 0
        yield self.clk.posedge
        self.re.next = 0
        self.wr_commit.next = 0
        self.wr_discard.next = 0
        self.rd_commit.next = 0
        self.rd_discard.next = 0

    def fifo_write_read(self, val, wcmd=None, rcmd=None):
        self.sfifo_model.write_read(val, wcmd, rcmd)
        self.we.next = 1
        self.din.next = val
        self.re.next = 1
        self.wr_commit.next = 1 if (wcmd == self.COMMIT  or wcmd == self.DISCARD_COMMIT) else 0
        self.wr_discard.next = 1 if (wcmd == self.DISCARD or wcmd == self.DISCARD_COMMIT) else 0
        self.rd_commit.next = 1 if (rcmd == self.COMMIT  or rcmd == self.DISCARD_COMMIT) else 0
        self.rd_discard.next = 1 if (rcmd == self.DISCARD or rcmd == self.DISCARD_COMMIT) else 0
        yield self.clk.posedge
        self.re.next = 0
        self.din.next = 0
        self.we.next = 0
        self.wr_commit.next = 0
        self.wr_discard.next = 0
        self.rd_commit.next = 0
        self.rd_discard.next = 0

    def fifo_state_check(self):
        yield delay(1)
        m = self.sfifo_model
        assert m.isFull()==self.full, "Full: expected={}, detected={}".format(m.isFull(), self.full)
        assert m.isEmpty()==self.empty, "Empty: expected={}, detected={}".format(m.isEmpty(), self.empty)
        assert m.isAFull()==self.afull, "AFull: expected={}, detected={}".format(m.isAFull(), self.afull)
        assert m.isAEmpty()==self.aempty, "AEmpty: expected={}, detected={}".format(m.isAEmpty(), self.aempty)
        assert m.getCount()==self.count, "Count: expected={}, detected={}".format(m.getCount(), self.count)
        assert m.getCountMax()==self.count_max, "CountMax: expected={}, detected={}".format(m.getCountMax(), self.count_max)
        assert m.isOvf()==self.ovf, "Overflow: expected={}, detected={}".format(m.isOvf(), self.ovf)
        assert m.isUdf()==self.udf, "Underflow: expected={}, detected={}".format(m.isUdf(), self.udf)
        if (not m.isEmpty() and not self.empty):
            assert m.getDout()==self.dout,"Dout: expected={}, detected={}".format(m.getDout(), self.dout)


    def testWriteThenReadCommitAtEveryOperation(self):
        ''' SFIFO:  Write then Read, Commit at every operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next(), wcmd=self.COMMIT)
                        yield self.fifo_state_check()
                    # Drain down to 0
                    for _ in range(1,i+1):
                        yield self.fifo_read(rcmd=self.COMMIT)
                        yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteThenReadCommitAtTheLastOperation(self):
        ''' SFIFO:  Write then Read, Commit at the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i
                    for w in range(1,i+1):
                        if (w != i):
                            yield self.fifo_write(data_in.next());
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_write(data_in.next(), wcmd=self.COMMIT)
                            yield self.fifo_state_check()
                    # Drain down to 0
                    for r in range(1,i+1):
                        if (r != i):
                            yield self.fifo_read();
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_read(rcmd=self.COMMIT)
                            yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteThenReadCommitAfterTheLastOperation(self):
        ''' SFIFO:  Write then Read, Commit after the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next());
                        yield self.fifo_state_check()
                    self.fifo_command(wcmd=self.COMMIT)
                    yield self.fifo_state_check()
                    # Drain down to 0
                    for _ in range(1,i+1):
                        yield self.fifo_read();
                        yield self.fifo_state_check()
                    self.fifo_command(rcmd=self.COMMIT)
                    yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteAndReadCommitAtEveryOperation(self):
        ''' SFIFO:  Write and Read simultaneously, Commit at every operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()
                yield self.fifo_write(data_in.next(), wcmd=self.COMMIT)
                yield self.fifo_state_check()

                for i in range(2,DEPTH+1):
                    for _ in range(1,i+1):
                        yield self.fifo_write_read(data_in.next(), wcmd=self.COMMIT, rcmd=self.COMMIT)
                        yield self.fifo_state_check()

                yield self.fifo_read(data_in.next(), rcmd=self.COMMIT)
                yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteAndReadCommitAtTheLastOperation(self):
        ''' SFIFO:  Write and Read simultaneously, Commit at the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()
                yield self.fifo_write(data_in.next(), wcmd=self.COMMIT);
                yield self.fifo_state_check()

                for i in range(2,DEPTH+1):
                    for w in range(1,i+1):
                        if (w != i):
                            yield self.fifo_write_read(data_in.next());
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_write_read(data_in.next(), wcmd=self.COMMIT, rcmd=self.COMMIT)
                            yield self.fifo_state_check()

                yield self.fifo_read(data_in.next(), rcmd=self.COMMIT);
                yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst


        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteAndReadCommitAfterTheLastOperation(self):
        ''' SFIFO:  Write and Read simultaneously, Commit after the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()
                yield self.fifo_write(data_in.next(), wcmd=self.COMMIT);
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next());
                        yield self.fifo_state_check()
                    self.fifo_command(wcmd=self.COMMIT)
                    yield self.fifo_state_check()
                    # Drain down to 0
                    for _ in range(1,i+1):
                        yield self.fifo_read();
                        yield self.fifo_state_check()
                    self.fifo_command(rcmd=self.COMMIT)
                    yield self.fifo_state_check()

                yield self.fifo_read(rcmd=self.COMMIT);
                yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteThenReadDiscardAtEveryOperation(self):
        ''' SFIFO:  Write then Read, Discard at every operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i and Discard
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next(), wcmd=self.DISCARD)
                        yield self.fifo_state_check()
                    # Fill up to i and Commit
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next(), wcmd=self.COMMIT)
                        yield self.fifo_state_check()
                    # Drain down to 0 and Discard
                    for _ in range(1,i+1):
                        yield self.fifo_read(rcmd=self.DISCARD)
                        yield self.fifo_state_check()
                    # Drain down to 0 and Commit
                    for _ in range(1,i+1):
                        yield self.fifo_read(rcmd=self.COMMIT)
                        yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteThenReadDiscardAtTheLastOperation(self):
        ''' SFIFO:  Write then Read, Discard at the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i and Discard
                    for w in range(1,i+1):
                        if (w != i):
                            yield self.fifo_write(data_in.next());
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_write(data_in.next(), wcmd=self.DISCARD)
                            yield self.fifo_state_check()
                    # Fill up to i and Commit
                    for w in range(1,i+1):
                        if (w != i):
                            yield self.fifo_write(data_in.next());
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_write(data_in.next(), wcmd=self.COMMIT)
                            yield self.fifo_state_check()
                    # Drain down to 0 and Discard
                    for r in range(1,i+1):
                        if (r != i):
                            yield self.fifo_read();
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_read(rcmd=self.DISCARD)
                            yield self.fifo_state_check()
                    # Drain down to 0 and Commit
                    for r in range(1,i+1):
                        if (r != i):
                            yield self.fifo_read();
                            yield self.fifo_state_check()
                        else:
                            yield self.fifo_read(rcmd=self.COMMIT)
                            yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testWriteThenReadDiscardAfterTheLastOperation(self):
        ''' SFIFO:  Write then Read, Discard after the last operation '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                yield self.reset()
                yield self.fifo_state_check()

                for i in range(1,DEPTH+1):
                    # Fill up to i and Discard
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next());
                        yield self.fifo_state_check()
                    self.fifo_command(wcmd=self.DISCARD)
                    yield self.fifo_state_check()
                    # Fill up to i and Commit
                    for _ in range(1,i+1):
                        yield self.fifo_write(data_in.next());
                        yield self.fifo_state_check()
                    self.fifo_command(wcmd=self.COMMIT)
                    yield self.fifo_state_check()
                    # Drain down to 0 and Discard
                    for _ in range(1,i+1):
                        yield self.fifo_read();
                        yield self.fifo_state_check()
                    self.fifo_command(rcmd=self.DISCARD)
                    yield self.fifo_state_check()
                    # Drain down to 0 and Commit
                    for _ in range(1,i+1):
                        yield self.fifo_read();
                        yield self.fifo_state_check()
                    self.fifo_command(rcmd=self.COMMIT)
                    yield self.fifo_state_check()

                yield self.clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                self.sfifo_model = sfifo_beh.sfifo_beh(dpt)
                dut = getDut( fifo_speculative,
                              rst = self.rst,
                              clk = self.clk,
                              full = self.full,
                              we = self.we,
                              din = self.din,
                              empty = self.empty,
                              re = self.re,
                              dout = self.dout,
                              wr_commit = self.wr_commit,
                              wr_discard = self.wr_discard,
                              rd_commit = self.rd_commit,
                              rd_discard = self.rd_discard,
                              afull = self.afull,
                              aempty = self.aempty,
                              count = self.count,
                              afull_th = None,
                              aempty_th = None,
                              ovf = self.ovf,
                              udf = self.udf,
                              count_max = self.count_max,
                              depth = dpt,
                              width = None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testSFifo']
    unittest.main()



import unittest

from myhdl import *
from myhdl_lib.utils import byteorder
import myhdl_lib.simulation as sim

import random


class TestByteorder(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def test56bits(self):
        ''' BYTEORDER: 7*8 bits'''
        BYTES = 7
        BITS = BYTES*8

        bv_di = Signal(intbv(0)[BITS:])
        bv_do = Signal(intbv(0)[BITS:])

        argl = {"bv_di":bv_di, "bv_do":bv_do, "REVERSE":True}

        def stim():
            @instance
            def _inst():
                bv_di.next = 0
                yield delay(10)
                assert 0 == bv_do, "Byteorder (reverse={}): expected {}, detected {}".format(reverse, 0, bv_do)

                for _ in range(10):
                    bv_di.next = random.randint(0,bv_di.max-1)
                    yield delay(10)
                    if reverse:
                        for b in range(BYTES):
                            assert bv_di[(b+1)*8:b*8] == bv_do[(BYTES-b)*8:(BYTES-b-1)*8], "Byteorder byte {}: {} {}".format(b, hex(bv_di,BYTES), hex(bv_do,BYTES))
                    else:
                        for b in range(BYTES):
                            assert bv_di[(b+1)*8:b*8] == bv_do[(b+1)*8:b*8], "Byteorder byte {}: {} {}".format(b, hex(bv_di), hex(bv_do))

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for reverse in [False, True]:
                argl["REVERSE"] = reverse
                dut=getDut(byteorder, **argl)
                stm = stim()
                Simulation( dut, stm).run()
                del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


from myhdl import *
from mem import ram_sdp_ar



def fifo_speculative(rst, clk, full, we, din, empty, re, dout, wr_commit=None, wr_discard=None, rd_commit=None, rd_discard=None, afull=None, aempty=None, count=None, afull_th=None, aempty_th=None, ovf=None, udf=None, count_max=None,  depth=None, width=None):
    """ Speculative FIFO

        Input  interface: full,  we, din,  wr_commit, wr_discard
        Output interface: empty, re, dout, rd_commit, rd_discard

        we         (i) - writes data speculatively (the data can be later removed from the fifo)
        wr_commit  (i) - accept the speculatively written data (can be read)
        wr_discard (i) - remove the speculatively written data at once
        full       (o) - asserted when there are no empty cells to write data speculatively

        If we is asserted together with the wr_commit or wr_discard, the data being written is affected by the commit/discard command.
        If wr_commit and wr_discard are asserted simultaneously, the wr_discard wins.

        re         (i) - reads data speculatively (the data can be later restored)
        rd_commit  (i) - removes the speculatively read data at once
        rd_discard (i) - restores the speculatively read data (will be read again)
        empty      (o) - asserted when there are no committed date to be read

        If re is asserted together with the rd_commit or rd_discards, the data being read is affected by the commit/discard command.
        If rd_commit and rd_discard are asserted simultaneously, the rd_discard wins.

        Note: This fifo can be Full and Empty at the same time - all fifo cells are occupied by data that are either speculatively written or speculatively read

        Extra interface:
        afull      (o) - almost full flag, asserted when the number of empty cells <= afull_th
        aempty     (o) - almost empty flag, asserted when the number of full cells <= aempty_th

        afull_th   (i) - almost full threshold, in terms of fifo cells; Optional, default depth/2
        aempty_th  (i) - almost empty threshold, in terms of fifo cells; Optional, default depth/2

        count      (o) - number of occupied fifo cells, committed and not committed (speculatively written/read)
        count_max  (o) - max number of occupied fifo cells reached since the last reset
        ovf        (o) - overflow flag, set at the first write in a full fifo, cleared at reset
        udf        (o) - underflow flag, set at the first read from an empty fifo, cleared at reset
    """

    if (width == None):
        width = len(din)
    if (depth == None):
        depth = 2

    assert ((wr_commit != None) and (wr_discard != None)) or ((wr_commit == None) and (wr_discard == None)), "SFIFO ERROR: Interface signals wr_commit and wr_discard must be either both used or both left unused"
    assert ((rd_commit != None) and (rd_discard != None)) or ((rd_commit == None) and (rd_discard == None)), "SFIFO ERROR: Interface signals rd_commit and rd_discard must be either both used or both left unused"

    wr_commit       = wr_commit  if (wr_commit  != None) else Signal(bool(1))
    wr_discard      = wr_discard if (wr_discard != None) else Signal(bool(0))
    rd_commit       = rd_commit  if (rd_commit  != None) else Signal(bool(1))
    rd_discard      = rd_discard if (rd_discard != None) else Signal(bool(0))

    full_flg        = Signal(bool(1))
    empty_flg       = Signal(bool(1))
    we_safe         = Signal(bool(0))
    re_safe         = Signal(bool(0))
    swr_non0        = Signal(bool(0))
    srd_non0        = Signal(bool(0))
    wr_commit_non0  = Signal(bool(0))
    rd_commit_non0  = Signal(bool(0))
    wr_discard_non0 = Signal(bool(0))
    rd_discard_non0 = Signal(bool(0))

    # Almost full/empty and their thresholds
    afull_flg       = Signal(bool(0))
    aempty_flg      = Signal(bool(0))
    if (afull_th == None):
        afull_th = depth//2
    if (aempty_th == None):
        aempty_th = depth//2

    afull           = afull  if (afull  != None) else Signal(bool(0))
    aempty          = aempty if (aempty != None) else Signal(bool(0))
    count           = count  if (count  != None) else Signal(intbv(0, min=0, max=depth + 1))

    ovf             = ovf       if (ovf       != None) else Signal(bool(0))
    udf             = udf       if (udf       != None) else Signal(bool(0))
    count_max       = count_max if (count_max != None) else Signal(intbv(0, min=0, max=depth+1))

    rd_ptr          = Signal(intbv(0, min=0, max=depth))
    wr_ptr          = Signal(intbv(0, min=0, max=depth))
    srd_ptr         = Signal(intbv(0, min=0, max=depth))
    srd_ptr_new     = Signal(intbv(0, min=0, max=depth))
    swr_ptr         = Signal(intbv(0, min=0, max=depth))
    swr_ptr_new     = Signal(intbv(0, min=0, max=depth))

    # Data written in the fifo (committed and not committed)
    data_count      = Signal(intbv(0, min=0,   max=depth + 1))
    data_count_max  = Signal(intbv(0, min=0,   max=depth + 1))

    @always_comb
    def out_assign():
        full.next           = full_flg
        empty.next          = empty_flg
        afull.next          = afull_flg
        aempty.next         = aempty_flg
        count.next          = data_count
        count_max.next      = data_count_max

    @always_comb
    def safe_read_write():
        we_safe.next        = we and not full_flg
        re_safe.next        = re and not empty_flg

    @always_comb
    def non_zero_commits():
        wr_commit_non0.next     = wr_commit and (swr_non0 or we_safe) and not wr_discard
        rd_commit_non0.next     = rd_commit and (srd_non0 or re_safe) and not rd_discard
        #
        wr_discard_non0.next    = wr_discard and (swr_non0 or we_safe)
        rd_discard_non0.next    = rd_discard and (srd_non0 or re_safe)

    """ ---------------------------- """
    """ - Write, Read, Full, Empty - """
    """ ---------------------------- """

    @always_comb
    def sptrs_new():
        """ Next value for the speculative pointers """

        if (wr_discard):
            swr_ptr_new.next     = wr_ptr
        elif (we_safe):
            swr_ptr_new.next     = ((swr_ptr + 1) % depth)
        else:
            swr_ptr_new.next     = swr_ptr

        if (rd_discard):
            srd_ptr_new.next     = rd_ptr
        elif (re_safe):
            srd_ptr_new.next     = ((srd_ptr + 1) % depth)
        else:
            srd_ptr_new.next     = srd_ptr

    @always(clk.posedge)
    def state_main():
        if (rst):
            wr_ptr.next     = 0
            rd_ptr.next     = 0
            swr_ptr.next    = 0
            srd_ptr.next    = 0
            swr_non0.next   = 0
            srd_non0.next   = 0
            full_flg.next   = 0
            empty_flg.next  = 1
        else:

            # Speculative Write pointer
            swr_ptr.next = swr_ptr_new

            # Speculative Read pointer
            srd_ptr.next = srd_ptr_new

            # Write pointer
            if (wr_commit): wr_ptr.next = swr_ptr_new

            # Read pointer
            if (rd_commit): rd_ptr.next = srd_ptr_new

            # Empty flag
            if (wr_commit_non0 or rd_discard_non0):
                empty_flg.next  = 0
            elif (re_safe and (srd_ptr_new == wr_ptr)):
                empty_flg.next  = 1

            # Full flag
            if (rd_commit_non0 or wr_discard_non0):
                full_flg.next   = 0
            elif (we_safe and (swr_ptr_new == rd_ptr)):
                full_flg.next   = 1

            # swr_non0 flag
            if (wr_commit or wr_discard):
                swr_non0.next   = 0
            elif (we_safe):
                swr_non0.next   = 1

            # srd_non0 flag
            if (rd_commit or rd_discard):
                srd_non0.next   = 0
            elif (re_safe):
                srd_non0.next   = 1


    """
    rd                        srd            wr                        swr            rd
    ^                         ^              v                         v              ^
    |#########################|%%%%%%%%%%%%%%|*************************|..............|
    |<------ srd_count ------>|<------------>|<------ swr_count ------>|<------------>|
    |<--------------- wr_count ------------->|<--------------- rd_count ------------->|
    |<----------------------------------- depth ------------------------------------->|
    """

    """ ------------------------------------ """
    """ - Almost_Full, Almost_Empty, Count - """
    """ ------------------------------------ """
    # count of the speculatively written data
    swr_count       = Signal(intbv(0, min=0,   max=depth + 1))
    # count of the speculatively read data
    srd_count       = Signal(intbv(0, min=0,   max=depth + 1))
    # count of the write committed data
    wr_count        = Signal(intbv(0, min=0,   max=depth + 1))
    # count of the not write committed positions
    rd_count        = Signal(intbv(0, min=0,   max=depth + 1))

    @always(clk.posedge)
    def state_extra():
        if (rst):
            swr_count.next      = 0
            srd_count.next      = 0
            wr_count.next       = 0
            rd_count.next       = depth
            data_count.next     = 0
            afull_flg.next      = 0
            aempty_flg.next     = 1
            data_count_max.next = 0
        else:
            swr_count_new   = 0
            if (not wr_commit and not wr_discard):
                swr_count_new   = int((swr_count + 1) if (we_safe) else swr_count)

            swr_count.next  = swr_count_new

            srd_count_new   = 0
            if (not rd_commit and not rd_discard):
                srd_count_new   = int((srd_count + 1 if (re_safe) else srd_count))

            srd_count.next  = srd_count_new

            wr_add_new = 0
            if (wr_commit and not wr_discard):
                wr_add_new  = int((swr_count + 1) if (we_safe) else swr_count)

            rd_add_new = 0
            if (rd_commit and not rd_discard):
                rd_add_new  = int((srd_count + 1) if (re_safe) else srd_count)

            wr_count_new    = wr_count + wr_add_new - rd_add_new
            rd_count_new    = rd_count - wr_add_new + rd_add_new

            wr_count.next   = wr_count_new
            rd_count.next   = rd_count_new

            aempty_flg.next = ((wr_count_new - srd_count_new) <= aempty_th)
            afull_flg.next  = ((rd_count_new - swr_count_new) <= afull_th )

            # Count data in the fifo
            data_count_new  = wr_count_new + swr_count_new
            data_count.next = data_count_new

            if (data_count_max < data_count_new): data_count_max.next = data_count_new

    ''' Overflow flag '''
    if (ovf != None):
        @always(clk.posedge)
        def ovf_proc():
            if (rst):
                ovf.next = 0
            else:
                if (we and full_flg ):
                    ovf.next = 1

    ''' Underflow flag '''
    if (udf != None):
        @always(clk.posedge)
        def udf_proc():
            if (rst):
                udf.next = 0
            else:
                if (re and empty_flg):
                    udf.next = 1

    """ ------------------- """
    """ - Memory instance - """
    """ ------------------- """
    mem_we      = Signal(bool(0))
    mem_addrw   = Signal(intbv(0, min=0, max=depth))
    mem_addrr   = Signal(intbv(0, min=0, max=depth))
    mem_di      = Signal(intbv(0)[width:0])
    mem_do      = Signal(intbv(0)[width:0])

    # RAM: Simple-Dual-Port, Asynchronous Read
    mem = ram_sdp_ar(   clk     = clk,
                        we      = mem_we,
                        addrw   = mem_addrw,
                        addrr   = mem_addrr,
                        di      = mem_di,
                        do      = mem_do )

    @always_comb
    def mem_connect():
        mem_we.next         = we_safe
        mem_addrw.next      = swr_ptr
        mem_addrr.next      = srd_ptr
        mem_di.next         = din
        dout.next           = mem_do

    return instances()


def convert(depth=256, width=8):

    rst, clk   = [Signal(bool(0)) for _ in range(2)]
    din, dout  = [Signal(intbv(0)[width:]) for _ in range(2)]

    re, we, empty, full, afull, aempty = [Signal(bool(0)) for _ in range(6)]
    wr_commit, wr_discard, rd_commit, rd_discard = [Signal(bool(0)) for _ in range(4)]
    count       = Signal(intbv(0, min=0, max=depth+1))
    afull_th    = Signal(intbv(0, min=0, max=depth+1))
    aempty_th   = Signal(intbv(0, min=0, max=depth+1))
    ovf         = Signal(bool(0))
    udf         = Signal(bool(0))
    count_max   = Signal(intbv(0, min=0, max=depth+1))

    toVerilog(fifo_speculative, rst, clk, full, we, din, empty, re, dout, wr_commit, wr_discard, rd_commit, rd_discard, afull, aempty, count, afull_th, aempty_th, ovf, udf, count_max, depth=depth, width=width)


if __name__ == "__main__":
    convert()



import unittest

from myhdl import *
from myhdl_lib.handshake import hs_arbdemux
import myhdl_lib.simulation as sim
from myhdl_lib.utils import assign



class TestArbDemux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    @staticmethod
    def hs_arbdemux_top(rst, clk, i_rdy, i_vld, o0_rdy, o0_vld, o1_rdy, o1_vld, o2_rdy, o2_vld, sel, ARBITER_TYPE):
        ''' Needed when hs_arbdemux is co-simulated as top level'''

        hsi_rdy = Signal(bool(0))
        hsi_vld = Signal(bool(0))
        hso0_rdy = Signal(bool(0))
        hso0_vld = Signal(bool(0))
        hso1_rdy = Signal(bool(0))
        hso1_vld = Signal(bool(0))
        hso2_rdy = Signal(bool(0))
        hso2_vld = Signal(bool(0))

        @always_comb
        def _assign():
            hso0_rdy.next = o0_rdy
            hso1_rdy.next = o1_rdy
            hso2_rdy.next = o2_rdy

            o0_vld.next = hso0_vld
            o1_vld.next = hso1_vld
            o2_vld.next = hso2_vld

            i_rdy.next = hsi_rdy
            hsi_vld.next = i_vld

        hsi = (hsi_rdy, hsi_vld)
        ls_hso = [(hso0_rdy, hso0_vld), (hso1_rdy, hso1_vld), (hso2_rdy, hso2_vld)]

        _inst = hs_arbdemux(clk=clk, rst=rst, hsi=hsi, ls_hso=ls_hso, sel=sel, ARBITER_TYPE=ARBITER_TYPE)

        return instances()


    def testArbDemux3Prio(self):
        ''' HS_ARBDEMUX: 3 outputs, priority arbiter '''
        NUM_OUTPUTS = 3

        hsi_rdy = Signal(bool(0))
        hsi_vld = Signal(bool(0))
        hsi = (hsi_rdy, hsi_vld)

        ls_hso_rdy = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso_vld = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso = zip(ls_hso_rdy, ls_hso_vld)

        sel = Signal(intbv(0)[NUM_OUTPUTS:])

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        argl = {"rst":None, "clk":None,
                "i_rdy":hsi[0], "i_vld":hsi[1],
                "o0_rdy":ls_hso[0][0], "o0_vld":ls_hso[0][1],
                "o1_rdy":ls_hso[1][0], "o1_vld":ls_hso[1][1],
                "o2_rdy":ls_hso[2][0], "o2_vld":ls_hso[2][1],
                "sel":sel, "ARBITER_TYPE":"priority"}

        def stim():
            @instance
            def _inst():
                for i in range(NUM_OUTPUTS):
                    ls_hso_rdy[i].next = 0
                hsi_vld.next = 0
                yield delay(10)
                for i in range(NUM_OUTPUTS):
                    assert 0==ls_hso_vld[i], "hso_vld[{}]: expected {}, detected {}".format(i, 0, ls_hso_vld[i])
                assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format( 0, hsi_rdy)

                x = intbv(0)[NUM_OUTPUTS+1:]
                for i in range(x.max):
                    x[:] = i
                    for k in range(NUM_OUTPUTS):
                        ls_hso_rdy[k].next = x[k]
                    hsi_vld.next = x[NUM_OUTPUTS]

                    yield delay(10)

                    s = 0
                    for i in range(NUM_OUTPUTS):
                        if x[i]==1:
                            s = i
                            break

                    assert s==sel, "sel: expecte {}, dectected {}".format(s, sel)

                    for k in range(NUM_OUTPUTS):
                        if (s==k) and (x[NUM_OUTPUTS]==1):
                            assert 1==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 1, ls_hso_vld[k])
                        else:
                            assert 0==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 0, ls_hso_vld[k])

                    if x[s]==1:
                        assert 1==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(1, hsi_rdy)
                    else:
                        assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(0, hsi_rdy)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut=getDut(self.hs_arbdemux_top, **argl)
            clkgen = clk.gen()
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


    def testArbDemux3RR(self):
        ''' HS_ARBDEMUX: 3 outputs, roundrobin arbiter '''
        NUM_OUTPUTS = 3

        hsi_rdy = Signal(bool(0))
        hsi_vld = Signal(bool(0))
        hsi = (hsi_rdy, hsi_vld)

        ls_hso_rdy = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso_vld = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso = zip(ls_hso_rdy, ls_hso_vld)

        sel = Signal(intbv(0)[NUM_OUTPUTS:])

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        argl = {"rst":rst, "clk":clk,
                "i_rdy":hsi[0], "i_vld":hsi[1],
                "o0_rdy":ls_hso[0][0], "o0_vld":ls_hso[0][1],
                "o1_rdy":ls_hso[1][0], "o1_vld":ls_hso[1][1],
                "o2_rdy":ls_hso[2][0], "o2_vld":ls_hso[2][1],
                "sel":sel, "ARBITER_TYPE":"roundrobin"}

        def stim():
            @instance
            def _inst():
                for i in range(NUM_OUTPUTS):
                    ls_hso_rdy[i].next = 0
                hsi_vld.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                for i in range(NUM_OUTPUTS):
                    assert 0==ls_hso_vld[i], "hso_vld[{}]: expected {}, detected {}".format(i, 0, ls_hso_vld[i])
                assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format( 0, hsi_rdy)

                prio = NUM_OUTPUTS-1
                x = intbv(0)[NUM_OUTPUTS+1:]
                for i in range(x.max):
                    x[:] = i
                    for k in range(NUM_OUTPUTS):
                        ls_hso_rdy[k].next = x[k]
                    hsi_vld.next = x[NUM_OUTPUTS]

                    yield clk.posedge

                    s = 0
                    for p in range(NUM_OUTPUTS):
                        j = (prio+p+1)%NUM_OUTPUTS
                        if x[j]==1:
                            s = j
                            if hsi_rdy and hsi_vld:
                                prio = j
                            break

                    assert s==sel, "sel: expecte {}, dectected {}".format(s, sel)

                    for k in range(NUM_OUTPUTS):
                        if (s==k) and (x[NUM_OUTPUTS]==1):
                            assert 1==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 1, ls_hso_vld[k])
                        else:
                            assert 0==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 0, ls_hso_vld[k])

                    if x[s]==1:
                        assert 1==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(1, hsi_rdy)
                    else:
                        assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(0, hsi_rdy)

                yield clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            dut=getDut(self.hs_arbdemux_top, **argl)
            clkgen = clk.gen()
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


__version__ = "0.1.0"

from myhdl_lib.mem import rom, ram_sp_rf, ram_sp_wf, ram_sp_ar,ram_sdp_rf, ram_sdp_wf, ram_sdp_ar, ram_dp_rf, ram_dp_wf, ram_dp_ar
from myhdl_lib.fifo import fifo
from myhdl_lib.fifo_speculative import fifo_speculative
from myhdl_lib.mux import mux, demux, ls_mux, ls_demux, bitslice_select, byteslice_select
from myhdl_lib.handshake import hs_join, hs_fork, hs_mux, hs_demux, hs_arbmux, hs_arbdemux
from myhdl_lib.arbiter import arbiter, arbiter_priority, arbiter_roundrobin
from myhdl_lib.pipeline_control import pipeline_control
from myhdl_lib.utils import assign, byteorder

__all__ = [
           "rom", "ram_sp_rf", "ram_sp_wf", "ram_sp_ar", "ram_sdp_rf", "ram_sdp_wf", "ram_sdp_ar", "ram_dp_rf", "ram_dp_wf", "ram_dp_ar",
           "fifo",
           "fifo_speculative",
           "mux", "demux", "ls_mux", "ls_demux", "bitslice_select", "byteslice_select",
           "hs_join", "hs_fork", "hs_mux", "hs_demux", "hs_arbmux", "hs_arbdemux",
           "arbiter", "arbiter_priority", "arbiter_roundrobin",
           "pipeline_control",
           "assign", "byteorder"
           ]




import os
import sys
import fileinput

from myhdl import toVerilog, toVHDL, Cosimulation, traceSignals


class DUTer(object):
    ''' Returns a simulation instance of a MyHDL function, intended as a DUT in testbench
        The user selects a simulator: MyHDL, or co-simulation with external HDL simulator, e.g. icarus
        The user selects whether traces are generated or not
    '''


    def __init__(self):
        '''  '''
        self._simulator = "myhdl"
        self._trace = False

        self.sim_reg = {}

        self.registerSimulator(
            name="icarus",
            hdl="Verilog",
            analyze_cmd='iverilog -o {topname}.o tb_{topname}.v {topname}.v',
            simulate_cmd="vvp -m ./myhdl.vpi {topname}.o"
        )

    def registerSimulator(self, name=None, hdl=None, analyze_cmd=None, elaborate_cmd=None, simulate_cmd=None):
        ''' Registers an HDL _simulator
                name - str, user defined name, used to identify this _simulator record
                hdl - str, case insensitive, (verilog, vhdl), the HDL to which the simulated MyHDL code will be converted
                analyze_cmd - str, system command that will be run to analyze the generated HDL
                elaborate_cmd - str, optional, system command that will be run after the analyze phase
                simulate_cmd - str, system command that will be run to simulate the analyzed and elaborated design
                Before execution of a command string the following substitutions take place:
                    {topname} is substituted with the name of the simulated MyHDL function
        '''
        if not isinstance(name, str) or (name.strip() == ""):
            raise ValueError("Invalid _simulator name")
        if hdl.lower() not in ("vhdl", "verilog"):
            raise ValueError("Invalid hdl {}".format(hdl))
        if not isinstance(analyze_cmd, str) or (analyze_cmd.strip() == ""):
            raise ValueError("Invalid analyzer command")
        if elaborate_cmd is not None:
            if not isinstance(elaborate_cmd, str) or (elaborate_cmd.strip() == ""):
                raise ValueError("Invalid elaborate_cmd command")
        if not isinstance(simulate_cmd, str) or (simulate_cmd.strip() == ""):
            raise ValueError("Invalid _simulator command")

        self.sim_reg[name] = (hdl.lower(), analyze_cmd, elaborate_cmd, simulate_cmd)

    def selectSimulator(self, simulatorName):
        if not simulatorName:
            raise ValueError("No _simulator specified")
        if not simulatorName=='myhdl' and not self.sim_reg.has_key(simulatorName):
            raise ValueError("Simulator {} is not registered".format(simulatorName))
        self._simulator = simulatorName

    def enableTrace(self):
        self._trace = True

    def disableTrace(self):
        self._trace = False

    def _getCosimulation(self, func, **kwargs):
        ''' Returns a co-simulation instance of func. 
            Uses the _simulator specified by self._simulator. 
            Enables traces if self._trace is True
                func - MyHDL function to be simulated
                kwargs - dict of func interface assignments: for signals and parameters
        '''
        vals = {}
        vals['topname'] = func.func_name
        vals['unitname'] = func.func_name.lower()
        hdlsim = self._simulator
        if not hdlsim:
            raise ValueError("No _simulator specified")
        if not self.sim_reg.has_key(hdlsim):
            raise ValueError("Simulator {} is not registered".format(hdlsim))

        hdl, analyze_cmd, elaborate_cmd, simulate_cmd = self.sim_reg[hdlsim]

        # Convert to HDL
        if hdl == "verilog":
            toVerilog(func, **kwargs)
            if self._trace:
                self._enableTracesVerilog("./tb_{topname}.v".format(**vals))
        elif hdl == "vhdl":
            toVHDL(func, **kwargs)

        # Analyze HDL
        os.system(analyze_cmd.format(**vals))
        # Elaborate
        if elaborate_cmd:
            os.system(elaborate_cmd.format(**vals))
        # Simulate
        return Cosimulation(simulate_cmd.format(**vals), **kwargs)


    def _enableTracesVerilog(self, verilogFile):
        ''' Enables traces in a Verilog file'''
        fname, _ = os.path.splitext(verilogFile)
        inserted = False
        for _, line in enumerate(fileinput.input(verilogFile, inplace = 1)):
            sys.stdout.write(line)
            if line.startswith("end") and not inserted:
                sys.stdout.write('\n\n') 
                sys.stdout.write('initial begin\n')
                sys.stdout.write('    $dumpfile("{}_cosim.vcd");\n'.format(fname)) 
                sys.stdout.write('    $dumpvars(0, dut);\n') 
                sys.stdout.write('end\n\n') 
                inserted = True


    def _getDut(self, func, **kwargs):
        ''' Returns a simulation instance of func. 
            Uses the simulator specified by self._simulator. 
            Enables traces if self._trace is True
                func - MyHDL function to be simulated
                kwargs - dict of func interface assignments: for signals and parameters
        '''
        if self._simulator=="myhdl":
            if not self._trace:
                sim_dut = func(**kwargs)
            else:
                sim_dut = traceSignals(func, **kwargs)
        else:
            sim_dut = self._getCosimulation(func, **kwargs)

        return sim_dut

    def __call__(self, func, **kwargs):
        return self._getDut(func, **kwargs)




from myhdl import *
from utils import assign

def pipeline_control(rst, clk, rx_rdy, rx_vld, tx_rdy, tx_vld, stage_enable, stop_rx=None, stop_tx=None):
    """ Synchronizes data flow in a processing pipeline. 
            rx_rdy, rx_vld,      - (o)(i) handshake at the pipeline input (front of the pipeline)
            tx_rdy, tx_vld,      - (i)(o) handshake at the pipeline output (back of the pipeline)
            stage_enable         - (o) vector of enable signals, one signal per stage, that controls the data registration in the stages;
                                   The length of this vector determines the number of stages in the pipeline
            stop_rx              - (i) optional, vector of signals, one signal per stage; when asserted, the corresponding stage stops consuming data;
                                   allows for multicycle execution in a stage (e.g. consume a data, then process it multiple cycles)
            stop_tx              - (i) optional, vector of signals, one signal per stage; when asserted, the corresponding stage stops producing data;
                                    allows for multicycle execution in a stage (consume multiple data to produce single data )

            stop_rx and stop_tx  - If you do not need them, then do not connect them
    """

    NUM_STAGES = len(stage_enable)

    if (stop_rx == None):
        stop_rx = Signal(intbv(0)[NUM_STAGES:])

    if (stop_tx == None):
        stop_tx = Signal(intbv(0)[NUM_STAGES:])

    assert (len(stop_rx)==NUM_STAGES), "pipeline_control: expects len(stop_rx)=len(stage_enable), but len(stop_rx)={} len(stage_enable)={}".format(len(stop_rx),NUM_STAGES)
    assert (len(stop_tx)==NUM_STAGES), "pipeline_control: expects len(stop_tx)=len(stage_enable), but len(stop_tx)={} len(stage_enable)={}".format(len(stop_tx),NUM_STAGES)

    vld_reg = Signal(intbv(0)[NUM_STAGES:])
    vld = Signal(intbv(0)[NUM_STAGES:])
    rdy = Signal(intbv(0)[NUM_STAGES:])
    rdy_s = Signal(intbv(0)[NUM_STAGES:])

    @always_seq(clk.posedge, reset=rst)
    def regs():
        for s in range(NUM_STAGES):
            if rdy[s]:
                vld_reg.next[s] = vld[s]

    @always_comb
    def comb_chain():
        vld.next[0] = rx_vld or stop_rx[0]
        rdy.next[NUM_STAGES-1] = tx_rdy or stop_tx[NUM_STAGES-1] or not vld_reg[NUM_STAGES-1]
        for i in range(1,NUM_STAGES):
            vld.next[i] = (vld_reg[i-1] and not stop_tx[i-1]) or stop_rx[i]
            rdy.next[NUM_STAGES-1-i] = (rdy_s[NUM_STAGES-1-i+1] and not stop_rx[NUM_STAGES-1-i+1]) or stop_tx[NUM_STAGES-1-i]

    @always_comb
    def comb_assign():
        tx_vld.next = vld_reg[NUM_STAGES-1] and not stop_tx[NUM_STAGES-1]
        rx_rdy.next = rdy[0] and not stop_rx[0]
        stage_enable.next = vld[NUM_STAGES:] & rdy[NUM_STAGES:]
        rdy_s.next = rdy


    return instances()



def pipeline_control_new(rst, clk, rx_rdy, rx_vld, tx_rdy, tx_vld, stage_enable, stop_rx=None, stop_tx=None):
    """ Pipeline control unit
            rx_rdy, rx_vld,      - (o)(i) handshake at the pipeline input (front of the pipeline)
            tx_rdy, tx_vld,      - (i)(o) handshake at the pipeline output (back of the pipeline)
            stage_enable         - (o) vector of enable signals, one signal per stage, that controls the data registration in the stages;
                                   The length of this vector determines the number of stages in the pipeline
            stop_rx              - (i) optional, vector of signals, one signal per stage; when asserted, the corresponding stage stops consuming data;
                                   allows for multicycle execution in a stage (e.g. consume a data, then process it multiple cycles)
            stop_tx              - (i) optional, vector of signals, one signal per stage; when asserted, the corresponding stage stops producing data;
                                    allows for multicycle execution in a stage (consume multiple data to produce single data )

            stop_rx and stop_tx  - If you do not need them, then do not connect them
    """

    NUM_STAGES = len(stage_enable)

    if (stop_rx == None):
        stop_rx = Signal(intbv(0)[NUM_STAGES:])

    if (stop_tx == None):
        stop_tx = Signal(intbv(0)[NUM_STAGES:])

    assert (len(stop_rx)==NUM_STAGES), "pipeline_control: expects len(stop_rx)=len(stage_enable), but len(stop_rx)={} len(stage_enable)={}".format(len(stop_rx),NUM_STAGES)
    assert (len(stop_tx)==NUM_STAGES), "pipeline_control: expects len(stop_tx)=len(stage_enable), but len(stop_tx)={} len(stage_enable)={}".format(len(stop_tx),NUM_STAGES)

    rdy = [Signal(bool(0)) for _ in range(NUM_STAGES+1)]
    vld = [Signal(bool(0)) for _ in range(NUM_STAGES+1)]
    BC = NUM_STAGES*[False]
    en = [Signal(bool(0)) for _ in range(NUM_STAGES)]
    stop_rx_s = [Signal(bool(0)) for _ in range(NUM_STAGES)]
    stop_tx_s = [Signal(bool(0)) for _ in range(NUM_STAGES)]

    rdy[0] = rx_rdy
    vld[0] = rx_vld
    rdy[-1] = tx_rdy
    vld[-1] = tx_vld
    BC[-1] = True

    stg = [None for _ in range(NUM_STAGES)]

    for i in range(NUM_STAGES):
        stg[i] = _stage_ctrl(rst = rst,
                             clk = clk,
                             rx_rdy = rdy[i],
                             rx_vld = vld[i],
                             tx_rdy = rdy[i+1],
                             tx_vld = vld[i+1],
                             stage_en = en[i],
                             stop_rx = stop_rx_s[i],
                             stop_tx = stop_tx_s[i],
                             BC = BC[i])

    x = en[0] if NUM_STAGES==1 else ConcatSignal(*reversed(en))

    @always_comb
    def _comb():
        stage_enable.next = x
        for i in range(NUM_STAGES):
            stop_rx_s[i].next = stop_rx[i]
            stop_tx_s[i].next = stop_tx[i]

    return instances()


def _stage_ctrl(rst, clk, rx_rdy, rx_vld, tx_rdy, tx_vld, stage_en, stop_rx=None, stop_tx=None, BC=False):
    ''' Single stage control
            BC - enable bubble compression
    '''
    if stop_rx==None:
        stop_rx = False

    if stop_tx==None:
        stop_tx = False

    state = Signal(bool(0))
    a = Signal(bool(0))
    b = Signal(bool(0))
    bc_link = state if BC else True

    @always_comb
    def _comb1():
        a.next = tx_rdy or stop_tx or not bc_link
        b.next = rx_vld or stop_rx

    @always_comb
    def _comb2():
        rx_rdy.next = a and not stop_rx
        tx_vld.next = state and not stop_tx
        stage_en.next = a and b

    @always_seq(clk.posedge, reset=rst)
    def _state():
        if a:
            state.next = b

    return _comb1, _comb2, _state










import unittest
import itertools

from myhdl import *
from myhdl_lib.fifo import fifo
import myhdl_lib.simulation as sim


class TestFifo(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def setUp(self):
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 128
        DEPTH_MAX = 101

        self.full = Signal(bool(0))
        self.we = Signal(bool(0))
        self.din = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))
        self.empty = Signal(bool(0))
        self.re = Signal(bool(0))
        self.dout = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))
        self.afull = Signal(bool(0))
        self.aempty = Signal(bool(0))
        self.afull_th = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.aempty_th = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.count = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.count_max = Signal(intbv(0, min=0, max=DEPTH_MAX))
        self.ovf = Signal(bool(0))
        self.udf = Signal(bool(0))

        self.clk = sim.Clock(val=0, period=10, units="ns")
        self.rst = sim.ResetSync(clk=self.clk, val=0, active=1)
        self.clkgen = self.clk.gen()


    def tearDown(self):
        pass


    def reset(self):
        yield self.rst.pulse(5)

    def fifo_write(self, val):
        self.we.next = 1
        self.din.next = val
        yield self.clk.posedge
        self.we.next = 0
        self.din.next = 0

    def fifo_read(self):
        self.re.next = 1
        yield self.clk.posedge
        self.re.next = 0

    def fifo_write_read(self,val):
        self.we.next = 1
        self.din.next = val
        self.re.next = 1
        yield self.clk.posedge
        self.re.next = 0
        self.din.next = 0
        self.we.next = 0

    def fifo_state_check(self, aFull, aEmpty, aDout=None, aAlmostFull=None, aAlmostEmpty=None, aCount=None,aCountMax=None, aOverflow=None, aUnderflow=None):
        yield delay(1)
        assert self.full==aFull, "Full: expected={}, detected={}".format(aFull, self.full)
        assert self.empty==aEmpty,"Empty: expected={}, detected={}".format(aEmpty, self.empty)
        if aDout!=None:
            assert self.dout==aDout,"Dout: expected={}, detected={}".format(aDout, self.dout)
        if aAlmostFull!=None:
            assert self.afull==aAlmostFull, "AFull: expected={}, detected={}".format(aAlmostFull, self.afull)
        if aAlmostEmpty!=None:
            assert self.aempty==aAlmostEmpty, "AEmpty: expected={}, detected={}".format(aAlmostEmpty, self.aempty)
        if aCount!=None:
            assert self.count==aCount, "Count: expected={}, detected={}".format(aCount, self.count)
        if aCountMax!=None:
            assert self.count_max==aCountMax, "CountMax: expected={}, detected={}".format(aCountMax, self.count_max)
        if aOverflow!=None:
            assert self.ovf==aOverflow, "Overflow: expected={}, detected={}".format(aOverflow, self.ovf)
        if aUnderflow!=None:
            assert self.udf==aUnderflow, "Underflow: expected={}, detected={}".format(aUnderflow, self.udf)


    def testEmptyFull(self):
        ''' FIFO: Empty and Full, up and down'''
        DEPTH = [1, 2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                for _ in range(3):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aOverflow=0, aUnderflow=0)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=data_out.next(), aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=None,
                              aempty=None,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testEmptyFull1(self):
        ''' FIFO: Empty and Full, circular (simultaneous Write and Read) '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                for _ in range(3):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aOverflow=0, aUnderflow=0)
                    # Cycle 
                    for _ in range(DEPTH+1):
                        yield self.fifo_write_read(data_in.next())
                        do = data_out.next()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=DEPTH-1, aOverflow=0, aUnderflow=0)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=data_out.next(), aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Cycle
                    for _ in range(DEPTH+1):
                        yield self.fifo_write_read(data_in.next())
                        do = data_out.next()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=1, aOverflow=0, aUnderflow=0)
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut(fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=None,
                              aempty=None,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testCount(self):
        ''' FIFO: Count and CountMax '''
        DEPTH = [7, 8, 9]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                cmax = 0
                for j in range(DEPTH):
                    # Fill up to j
                    do = data_out.next()
                    for i in range(j):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aCountMax=cmax, aOverflow=0, aUnderflow=0)
                    yield self.fifo_write(data_in.next())
                    cmax += 1
                    full = (j+1==DEPTH)
                    yield self.fifo_state_check(aFull=full, aEmpty=0, aDout=do, aCount=j+1, aCountMax=cmax, aOverflow=0, aUnderflow=0)

                    if not full:
                        # Cycle
                        for _ in range(j+1):
                            yield self.fifo_write_read(data_in.next())
                            do = data_out.next()
                            yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=j+1, aCountMax=cmax, aOverflow=0, aUnderflow=0)
                    # Drain down to 0
                    for i in range(j,0,-1):
                        yield self.fifo_read()
                        do = data_out.next()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i, aCountMax=cmax, aOverflow=0, aUnderflow=0)
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=cmax, aOverflow=0, aUnderflow=0)

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut(fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=None,
                              aempty=None,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt, width=None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testOverflowUnderflow(self):
        ''' FIFO: Overflow and Underflow '''
        DEPTH = [1, 2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                # Fill up to DEPTH-1
                do = data_out.next()
                for i in range(DEPTH-1):
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aOverflow=0, aUnderflow=0)
                # Fill up to DEPTH
                yield self.fifo_write(data_in.next())
                yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aOverflow=0, aUnderflow=0)
                # Overflow
                for _ in range(2):
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aCountMax=DEPTH, aOverflow=1, aUnderflow=0)
                # Drain down to 1
                for i in range(DEPTH-1):
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=data_out.next(), aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=1, aUnderflow=0)
                # Drain to 0
                yield self.fifo_read()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=1, aUnderflow=0)
                # Underflow
                for _ in range(2):
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=1, aUnderflow=1)
                    data_out.next()
                # Write, Read (to move the pointers one step further)
                yield self.fifo_write(data_in.next())
                yield self.fifo_read()
                data_out.next()

                # Fill up to DEPTH-1
                do = data_out.next()
                for i in range(DEPTH-1):
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aOverflow=1, aUnderflow=1)
                # Fill up to DEPTH
                yield self.fifo_write(data_in.next())
                yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aOverflow=1, aUnderflow=1)
                # Overflow
                for _ in range(2):
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aCountMax=DEPTH, aOverflow=1, aUnderflow=1)
                # Drain down to 1
                for i in range(DEPTH-1):
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=data_out.next(), aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=1, aUnderflow=1)
                # Drain to 0
                yield self.fifo_read()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=1, aUnderflow=1)
                # Underflow
                for _ in range(2):
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=1, aUnderflow=1)
                    data_out.next()
                # Write, Read (to move the pointers one step further)
                yield self.fifo_write(data_in.next())
                yield self.fifo_read()
                data_out.next()


                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut(fifo,
                             rst=self.rst,
                             clk=self.clk,
                             full=self.full,
                             we=self.we,
                             din=self.din,
                             empty=self.empty,
                             re=self.re,
                             dout=self.dout,
                             afull=None,
                             aempty=None,
                             count=self.count,
                             afull_th=None,
                             aempty_th=None,
                             ovf=self.ovf,
                             udf=self.udf,
                             count_max=self.count_max,
                             depth=dpt,
                             width=None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testAlmostEmptyAlmostFullDefaultThreshold(self):
        ''' FIFO: AlmostEmpty and AlmostFull with default thresholds '''
        DEPTH = [1, 2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            DEFAULTH_THRESHOLD = DEPTH//2
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0, aAlmostEmpty=1)

                for _ in range(3):
                    # Fill up to DEPTH
                    do = data_out.next()
                    for i in range(1,DEPTH+1):
                        full = (i==DEPTH)
                        aempty = (i<=DEFAULTH_THRESHOLD)
                        afull = (i>=DEPTH-DEFAULTH_THRESHOLD)
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=full, aEmpty=0, aDout=do, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                    # Drain down to 0
                    for i in range(DEPTH-1, -1,-1):
                        aempty = (i<=DEFAULTH_THRESHOLD)
                        afull = (i>=DEPTH-DEFAULTH_THRESHOLD)
                        yield self.fifo_read()
                        if (i!=0):
                            do = data_out.next()
                            yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                        else:
                            yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut(fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=self.afull,
                              aempty=self.aempty,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testAlmostEmptyAlmostFullConfigureThreshold(self):
        ''' FIFO: AlmostEmpty and AlmostFull with programmed thresholds '''
        DEPTH = [2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0, aAlmostEmpty=1)

                for THRESHOLD in [0, 1, DEPTH//2-1, DEPTH//2+1, DEPTH-1, DEPTH]:
                    AE_THRESHOLD = THRESHOLD
                    AF_THRESHOLD = THRESHOLD
                    self.aempty_th.next = AE_THRESHOLD
                    self.afull_th.next = AF_THRESHOLD
                    yield self.clk.posedge
                    # Fill up to DEPTH
                    do = data_out.next()
                    for i in range(1,DEPTH+1):
                        full = (i==DEPTH)
                        aempty = (i<=AE_THRESHOLD)
                        afull = (i>=DEPTH-AF_THRESHOLD)
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=full, aEmpty=0, aDout=do, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                    # Drain down to 0
                    for i in range(DEPTH-1, -1,-1):
                        aempty = (i<=AE_THRESHOLD)
                        afull = (i>=DEPTH-AF_THRESHOLD)
                        yield self.fifo_read()
                        if (i!=0):
                            do = data_out.next()
                            yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                        else:
                            yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=i, aAlmostEmpty=aempty, aAlmostFull=afull)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut(fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=self.afull,
                              aempty=self.aempty,
                              count=self.count,
                              afull_th=self.afull_th,
                              aempty_th=self.aempty_th,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None)
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()


    def testDataWidth1(self):
        ''' FIFO: Data width 1'''
        DEPTH = [1, 2, 4, 7, 8, 9, 10]
        self.din = Signal(bool(0)) # <-- Bool
        self.dout = Signal(bool(0)) # <-- Bool

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(2))
                data_out = itertools.cycle(xrange(2))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                for _ in range(3):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=do, aCount=i+1, aOverflow=0, aUnderflow=0)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=do, aCount=DEPTH, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=data_out.next(), aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=self.din,
                              empty=self.empty,
                              re=self.re,
                              dout=self.dout,
                              afull=None,
                              aempty=None,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


    def testDatawidth0(self):
        ''' FIFO: Data width 0 '''
        DEPTH = [1, 2, 4, 7, 8, 9, 10]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(2))
                data_out = itertools.cycle(xrange(2))

                yield self.reset()
                yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=0, aOverflow=0, aUnderflow=0)

                for _ in range(3):
                    # Fill up to DEPTH-1
#                     do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=None, aCount=i+1, aOverflow=0, aUnderflow=0)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(aFull=1, aEmpty=0, aDout=None, aCount=DEPTH, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(aFull=0, aEmpty=0, aDout=None, aCount=DEPTH-i-1, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(aFull=0, aEmpty=1, aCount=0, aCountMax=DEPTH, aOverflow=0, aUnderflow=0)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_read()
                    data_out.next()

                yield self.clk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo,
                              rst=self.rst,
                              clk=self.clk,
                              full=self.full,
                              we=self.we,
                              din=None, # <-- Not connected
                              empty=self.empty,
                              re=self.re,
                              dout=None, # <-- Not connected
                              afull=None,
                              aempty=None,
                              count=self.count,
                              afull_th=None,
                              aempty_th=None,
                              ovf=self.ovf,
                              udf=self.udf,
                              count_max=self.count_max,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.clkgen, dut, stm).run()
                del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testFifo']
    unittest.main()




import unittest
import itertools

from myhdl import *
from myhdl_lib.fifo_async import fifo_async
import myhdl_lib.simulation as sim


class TestFifoAsync(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def setUp(self):
        DATA_RANGE_MIN = 0
        DATA_RANGE_MAX = 128
        DEPTH_MAX = 101

        self.wfull = Signal(bool(0))
        self.we = Signal(bool(0))
        self.wdata = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))
        self.rempty = Signal(bool(0))
        self.re = Signal(bool(0))
        self.rdata = Signal(intbv(0, min=DATA_RANGE_MIN, max=DATA_RANGE_MAX))

        self.wclk = sim.Clock(val=0, period=7, units="ns")
        self.rclk = sim.Clock(val=0, period=11, units="ns")
        self.wrst = sim.ResetSync(clk=self.wclk, val=0, active=1)
        self.rrst = sim.ResetSync(clk=self.rclk, val=0, active=1)
        self.wclkgen = self.wclk.gen()
        self.rclkgen = self.rclk.gen()

    def tearDown(self):
        pass

    def wreset(self):
        yield self.wrst.pulse(5)

    def rreset(self):
        yield self.rrst.pulse(5)

    def fifo_write(self, val):
        yield self.wclk.posedge
        self.we.next = 1
        self.wdata.next = val
        yield self.wclk.posedge
        self.we.next = 0
        self.wdata.next = 0

    def fifo_read(self):
        yield self.rclk.posedge
        self.re.next = 1
        yield self.rclk.posedge
        self.re.next = 0

    def fifo_state_check(self, WFull, REmpty, RData=None):
        yield delay(1)
        if  WFull is not None:
            assert self.wfull==WFull, "Full: expected={}, detected={}".format(WFull, self.wfull)
        if REmpty is not None:
            assert self.rempty==REmpty,"Empty: expected={}, detected={}".format(REmpty, self.rempty)
        if RData!=None:
            assert self.rdata==RData,"Dout: expected={}, detected={}".format(RData, self.rdata)


    def testEmptyFull(self):
        ''' FIFO_Async: Empty and Full, up and down'''
        DEPTH = [4, 8, 16]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.rreset()
                yield self.wreset()
                yield self.fifo_state_check(WFull=0, REmpty=1)

                for _ in range(5):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(WFull=0, REmpty=None, RData=do)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(WFull=1, REmpty=None, RData=do)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(WFull=None, REmpty=0, RData=data_out.next())
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(WFull=0, REmpty=1)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    for _ in range(3): yield self.rclk.posedge # Waith for clock domain crossing sync and update of empty
                    yield self.fifo_read()
                    data_out.next()

                yield self.wclk.posedge
                yield self.rclk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo_async,
                              wrst=self.wrst,
                              rrst=self.rrst,
                              wclk=self.wclk,
                              rclk=self.rclk,
                              wfull=self.wfull,
                              we=self.we,
                              wdata=self.wdata,
                              rempty=self.rempty,
                              re=self.re,
                              rdata=self.rdata,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.wclkgen, self.rclkgen, dut, stm).run()
                del dut, stm


    def testOverflowUnderflow(self):
        ''' FIFO_Async: Does not write when full, does not read when empty'''
        DEPTH = [4, 8, 16]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.rreset()
                yield self.wreset()
                yield self.fifo_state_check(WFull=0, REmpty=1)

                for _ in range(5):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(WFull=0, REmpty=None, RData=do)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(WFull=1, REmpty=None, RData=do)
                    # Write when full
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(WFull=1, REmpty=None, RData=do)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(WFull=None, REmpty=0, RData=data_out.next())
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(WFull=0, REmpty=1)
                    # Read when empty
                    yield self.fifo_read()
                    yield self.fifo_state_check(WFull=0, REmpty=1)
                    data_out.next() # Remove the data written when full
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    for _ in range(3): yield self.rclk.posedge # Waith for clock domain crossing sync and update of empty
                    yield self.fifo_read()
                    data_out.next()

                yield self.wclk.posedge
                yield self.rclk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo_async,
                              wrst=self.wrst,
                              rrst=self.rrst,
                              wclk=self.wclk,
                              rclk=self.rclk,
                              wfull=self.wfull,
                              we=self.we,
                              wdata=self.wdata,
                              rempty=self.rempty,
                              re=self.re,
                              rdata=self.rdata,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.wclkgen, self.rclkgen, dut, stm).run()
                del dut, stm


    def testDataWidth1(self):
        ''' FIFO_Async: Data width 1'''
        DEPTH = [4, 8, 16]
        self.wdata = Signal(bool(0)) # <-- Bool
        self.rdata = Signal(bool(0)) # <-- Bool

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(2))
                data_out = itertools.cycle(xrange(2))

                yield self.rreset()
                yield self.wreset()
                yield self.fifo_state_check(WFull=0, REmpty=1)

                for _ in range(5):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(WFull=0, REmpty=None, RData=do)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(WFull=1, REmpty=None, RData=do)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(WFull=None, REmpty=0, RData=data_out.next())
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(WFull=0, REmpty=1)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    for _ in range(3): yield self.rclk.posedge # Waith for clock domain crossing sync and update of empty
                    yield self.fifo_read()
                    data_out.next()

                yield self.wclk.posedge
                yield self.rclk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo_async,
                              wrst=self.wrst,
                              rrst=self.rrst,
                              wclk=self.wclk,
                              rclk=self.rclk,
                              wfull=self.wfull,
                              we=self.we,
                              wdata=self.wdata,
                              rempty=self.rempty,
                              re=self.re,
                              rdata=self.rdata,
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.wclkgen, self.rclkgen, dut, stm).run()
                del dut, stm

    def testDatawidth0(self):
        ''' FIFO_Async: Data width 0 '''
        DEPTH = [4, 8, 16]

        def stim(DEPTH):
            @instance
            def _inst():
                data_in = itertools.cycle(xrange(128))
                data_out = itertools.cycle(xrange(128))

                yield self.rreset()
                yield self.wreset()
                yield self.fifo_state_check(WFull=0, REmpty=1)

                for _ in range(5):
                    # Fill up to DEPTH-1
                    do = data_out.next()
                    for i in range(DEPTH-1):
                        yield self.fifo_write(data_in.next())
                        yield self.fifo_state_check(WFull=0, REmpty=None, RData=None)
                    # Fill up to DEPTH
                    yield self.fifo_write(data_in.next())
                    yield self.fifo_state_check(WFull=1, REmpty=None, RData=None)
                    # Drain down to 1
                    for i in range(DEPTH-1):
                        yield self.fifo_read()
                        yield self.fifo_state_check(WFull=None, REmpty=0, RData=None)
                    # Drain to 0
                    yield self.fifo_read()
                    yield self.fifo_state_check(WFull=0, REmpty=1)
                    # Write, Read (to move the pointers one step further)
                    yield self.fifo_write(data_in.next())
                    for _ in range(3): yield self.rclk.posedge # Waith for clock domain crossing sync and update of empty
                    yield self.fifo_read()
                    data_out.next()

                yield self.wclk.posedge
                yield self.rclk.posedge

                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)
            for dpt in DEPTH:
                dut = getDut( fifo_async,
                              wrst=self.wrst,
                              rrst=self.rrst,
                              wclk=self.wclk,
                              rclk=self.rclk,
                              wfull=self.wfull,
                              we=self.we,
                              wdata=None, # <-- Not connected
                              rempty=self.rempty,
                              re=self.re,
                              rdata=None, # <-- Not connected
                              depth=dpt,
                              width=None
                            )
                stm = stim(dpt)
                Simulation(self.wclkgen, self.rclkgen, dut, stm).run()
                del dut, stm



if __name__ == '__main__':
    #import sys;sys.argv = ['', 'Test.testFifo']
    unittest.main()



import unittest

from myhdl import *
from myhdl_lib.mux import byteslice_select
import myhdl_lib.simulation as sim

import random


class TestBytesliceSelect(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]


    def test56to24(self):
        ''' BYTESLICE_SELECT: 7*8 bit -> 3*8 bit'''
        LEN_I = 7*8
        LEN_O = 3*8
        MAX_OFFSET = (LEN_I - LEN_O)//8

        bv_di = Signal(intbv(0)[LEN_I:])
        bv_do = Signal(intbv(0)[LEN_O:])
        offset = Signal(intbv(0, min=0, max=MAX_OFFSET+1+1))

        argl = {"offset":offset, "bv_di":bv_di, "bv_do":bv_do}

        def stim():
            @instance
            def _inst():
                offset.next = 0
                bv_di.next = 0
                yield delay(10)
                assert 0 == bv_do, "Byteslice_select (offset={}): expected {}, detected {}".format(offset, 0, bv_do)

                for _ in range(10):
                    bv_di.next = random.randint(0, bv_di.max)
                    for i in range(MAX_OFFSET+1+1):
                        offset.next = i
                        yield delay(10)
                        exp_out = bv_di[i*8+LEN_O:i*8]
                        if i<=MAX_OFFSET:
                            assert exp_out == bv_do, "Byteslice_select (offset={}): expected {}, detected {}".format(offset, exp_out, bv_do)
                        else:
                            assert 0 == bv_do, "Byteslice_select (offset={}): expected {}, detected {}".format(offset, 0, bv_do)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(byteslice_select, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest

from myhdl import *
from myhdl_lib.arbiter import arbiter_roundrobin
import myhdl_lib.simulation as sim



class TestArbiterRoundrobin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testOneHot(self):
        ''' ARBITER_ROUNDROBIN: One Hot requests'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))
        gnt_rdy = Signal(bool(0))

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                gnt_rdy.next = 0
                req_vec.next = 0
                yield rst.pulse(10)
                check(vec=0, idx=0, vld=0)

                gnt_rdy.next = 1
                for _ in range(3):
                    for i in range(NUM_REQ):
                        req_vec.next = 1 << i
                        yield delay(1)
                        check(vec=1<<i, idx=i, vld=1)
                        yield clk.posedge

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            clkgen = clk.gen()
            dut = getDut(arbiter_roundrobin, rst=rst, clk=clk, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld, gnt_rdy=gnt_rdy)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


    def testCount(self):
        ''' ARBITER_ROUNDROBIN: Count '''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))
        gnt_rdy = Signal(bool(0))

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                gnt_rdy.next = 0
                req_vec.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                check(vec=0, idx=0, vld=0)

                gnt_rdy.next = 1
                prio = NUM_REQ-1
                x = intbv(0)[NUM_REQ:]
                for i in range(x.max):
                    x[:] = i
                    req_vec.next = x
                    s = 0
                    v = 0
                    for k in range(NUM_REQ):
                        kk = (prio + k + 1)%NUM_REQ
                        if x[kk]==1:
                            s = kk
                            v = 1
                            break

                    yield delay(1)
                    check(vec=(1<<s)*v, idx=s, vld=v)

                    yield clk.posedge

                    if gnt_rdy==1:
                            prio = s

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            clkgen = clk.gen()
            dut = getDut(arbiter_roundrobin, rst=rst, clk=clk, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld, gnt_rdy=gnt_rdy)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


    def testAll(self):
        ''' ARBITER_ROUNDROBIN: All requests'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))
        gnt_rdy = Signal(bool(0))

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                gnt_rdy.next = 0
                req_vec.next = 0
                yield rst.pulse(10)
                check(vec=0, idx=0, vld=0)

                mask = req_vec.max-1
                req_vec.next = mask
                gnt_rdy.next = 1
                for _ in range(3):
                    for i in range(NUM_REQ):
                        yield delay(1)
                        check(vec=1<<i, idx=i, vld=1)
                        yield clk.posedge

                yield clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            clkgen = clk.gen()
            dut = getDut(arbiter_roundrobin, rst=rst, clk=clk, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld, gnt_rdy=gnt_rdy)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm


    def testAllDummy(self):
        ''' ARBITER_ROUNDROBIN: All requests, dummy strobe'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))
        gnt_rdy = Signal(bool(0))

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                gnt_rdy.next = 0
                req_vec.next = 0
                yield rst.pulse(10)
                check(vec=0, idx=0, vld=0)

                mask = req_vec.max-1
                req_vec.next = mask
                gnt_rdy.next = 1
                for _ in range(3):
                    for i in range(NUM_REQ):
                        # gnt_vld & gnt_rdy
                        req_vec.next = mask
                        yield delay(1)
                        check(vec=1<<i, idx=i, vld=1)
                        yield clk.posedge
                        # /gnt_vld & gnt_rdy
                        req_vec.next = 0
                        yield delay(1)
                        check(vec=0, idx=0, vld=0)
                        yield clk.posedge

                yield clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            clkgen = clk.gen()
            dut = getDut(arbiter_roundrobin, rst=rst, clk=clk, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld, gnt_rdy=gnt_rdy)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm

    def testAllGap(self):
        ''' ARBITER_ROUNDROBIN: All requests, gappy strobe'''
        NUM_REQ = 5
        req_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_vec = Signal(intbv(0)[NUM_REQ:])
        gnt_idx = Signal(intbv(0, min=0, max=NUM_REQ))
        gnt_vld = Signal(bool(0))
        gnt_rdy = Signal(bool(0))

        clk = sim.Clock(val=0, period=10, units="ns")
        rst = sim.ResetSync(clk=clk, val=0, active=1)

        def check(vec, idx, vld):
            assert vec == gnt_vec, "gnt_vec: expected {}, detected {}".format(vec, gnt_vec)
            assert idx == gnt_idx, "gnt_idx: expected {}, detected {}".format(idx, gnt_idx)
            assert vld == gnt_vld, "gnt_vld: expected {}, detected {}".format(vld, gnt_vld)

        def stim():
            @instance
            def _inst():
                gnt_rdy.next = 0
                req_vec.next = 0
                yield rst.pulse(10)
                yield clk.posedge
                check(vec=0, idx=0, vld=0)

                mask = req_vec.max-1
                req_vec.next = mask
                for i in range(NUM_REQ):
                    for _ in range(3):
                        yield delay(1)
                        check(vec=1<<i, idx=i, vld=1)
                        gnt_rdy.next = 0
                        yield clk.posedge
                    else:
                        yield delay(1)
                        check(vec=1<<i, idx=i, vld=1)
                        gnt_rdy.next = 1
                        yield clk.posedge

                yield clk.posedge
                gnt_rdy.next = 0
                yield clk.posedge
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            clkgen = clk.gen()
            dut = getDut(arbiter_roundrobin, rst=rst, clk=clk, req_vec=req_vec, gnt_vec=gnt_vec, gnt_idx=gnt_idx, gnt_vld=gnt_vld, gnt_rdy=gnt_rdy)
            stm = stim()
            Simulation(clkgen, dut, stm).run()
            del clkgen, dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


""" setuptools distribution and installation script. """

from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup( name                 = 'myhdl_lib',
       version              = '0.1.0',
       description          = 'TODO: Library of components based on Python and MyHDL',
       long_description     = readme(),
       classifiers          = [
           'Development Status :: 3 - Alpha',
           'License :: OSI Approved :: MIT License',
           'Programming Language :: Python :: 2.7',
           'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
       ],
       keywords             = 'myhdl pihdf fpga',
       url                  = 'https://github.com/nkavaldj/myhdl_lib.git',
       author               = 'Nikolay Kavaldjiev',
       author_email         = 'nikolay.kavaldjiev@gmail.com',
       license              = 'MIT',
       packages             = [
           'myhdl_lib',
           'myhdl_lib.simulation'
       ],
       install_requires     = [
           'myhdl'
       ],
       include_package_data = True,
       zip_safe             = False
    )



from myhdl import *

from math import log, floor, ceil
from myhdl_lib.utils import assign

def bytecount(rst, clk, rx_vld, rx_sop, rx_eop, rx_dat, rx_mty, count, MAX_BYTES=1500 ):
    """ Counts bytes in a stream of packetised data
            rx_vld - (i) valid data
            rx_sop - (i) start of packet
            rx_eop - (i) end of packet
            rx_dat - (i) data
            rx_mty - (i) empty bits when rx_eop
            count  - (o) byte count
            MAX_BYTES - a limit: maximum number of bytes that may arrive in a single packet
        The result is ready in the first clock cycle after rx_eop
    """

    DATA_WIDTH    = len(rx_dat) 

    assert DATA_WIDTH%8==0, "bytecount: expects len(rx_dat)=8*x, but len(rx_dat)={}".format(DATA_WIDTH)

    NUM_BYTES       = DATA_WIDTH // 8
    MAX_BYTES_LG    = floor(log(MAX_BYTES,2))
    MIN_COUNT_WIDTH = int(MAX_BYTES_LG) + 1

    assert len(count)>=MIN_COUNT_WIDTH, "bytecount: expects len(count)>={}, but len(count)={}".format(MIN_COUNT_WIDTH, len(count))

    count_s = Signal(intbv(0)[MIN_COUNT_WIDTH:])

    @always_seq(clk.posedge, reset=rst)
    def _count():
        if (rx_vld):
            x = NUM_BYTES
            if (rx_eop==1):
                x = x -rx_mty
            if (rx_sop==1):
                count_s.next = x
            else:
                count_s.next = count_s + x

    @always_comb
    def out_comb():
        count.next = count_s

    return instances()


def checksum(rst, clk, rx_vld, rx_sop, rx_eop, rx_dat, rx_mty, chksum, sum16=None, sumN=None, init_sum=None, MAX_BYTES=1500):
    """ Calculates checksum on a stream of packetised data
            rx_vld   - (i) valid data
            rx_sop   - (i) start of packet
            rx_eop   - (i) end of packet
            rx_dat   - (i) data
            rx_mty   - (i) empty bits when rx_eop
            init_sum - (i) initial value for the sum (e.g. sum of the some fields from a packet header)
            sumN     - (o) optional, plane sum
            sum16    - (o) optional, 16 bit sum (every time the sum overflows, the overflow is added to the sum)
            chksum   - (o) optional, checksum (~sum16)
            MAX_BYTES - a limit: maximum number of bytes that may arrive in a single packet
        Assumes Big-endian data.
        The results are ready in the first clock cycle after rx_eop
    """

    DATA_WIDTH = len(rx_dat) 

    assert DATA_WIDTH%16==0, "checksum: expects len(rx_dat)=16*x, but len(tx_dat)={}".format(DATA_WIDTH)

    NUM_BYTES = DATA_WIDTH // 8
    NUM_WORDS = DATA_WIDTH // 16

    SUM_WIDTH    = 16 + int(ceil(log(MAX_BYTES/2,2))) + 1
    SUM16_WIDTH  = 16 + int(ceil(log(NUM_WORDS,2))) + 1


    if (init_sum != None):
        assert len(init_sum)<=16, "checksum: expects len(init_sum)<={}, but len(init_sum)={}".format(16, len(init_sum))
        SUM_WIDTH    += 1
        SUM16_WIDTH  += 1
    else:
        init_sum = 0

    FF_MASK = intbv(0)[DATA_WIDTH:].max-1
    mdata = Signal(intbv(0)[DATA_WIDTH:])

    @always_comb
    def _mask_empty():
        mdata.next = rx_dat
        if rx_eop:
            mdata.next = rx_dat & (FF_MASK<<(rx_mty*8))

    # Slice masked data in 16 bit words
    mdata16 = [Signal(intbv(0)[16:]) for _ in range(NUM_WORDS)]
    _ass = [assign(mdata16[w], mdata((w+1)*16, w*16)) for w in range(NUM_WORDS)]

    if sumN!=None:
        assert len(sumN)>=SUM_WIDTH, "checksum: expects len(sumN)>={}, but len(sumN)={}".format(SUM_WIDTH, len(sumN))

        sumN_reg = Signal(intbv(0)[SUM_WIDTH:])

        @always_seq(clk.posedge, reset=rst)
        def _accuN():
            ''' Accumulate '''
            if (rx_vld):
                s = 0

                if (rx_sop):
                    s = int(init_sum)
                else:
                    s = int(sumN_reg)

                for w in range(NUM_WORDS):
                    s += mdata16[w]

                sumN_reg.next = s

        @always_comb
        def _passN():
            sumN.next = sumN_reg


    if chksum!=None or sum16!=None:

        sum16_reg = Signal(intbv(0)[SUM16_WIDTH:])

        @always_seq(clk.posedge, reset=rst)
        def _accu16():
            ''' Accumulate 16 bit words'''
            if (rx_vld):
                s = 0

                if (rx_sop):
                    s = int(init_sum)
                else:
                    s = int(sum16_reg)

                for w in range(NUM_WORDS):
                    s += mdata16[w]

                ss = intbv(s)[SUM16_WIDTH:]

                sum16_reg.next = ss[:2*8] + ss[2*8:]

        if sum16!=None:
            assert len(sum16)>=16, "checksum: expects len(sum16)>={}, but len(sum16)={}".format(16, len(sum16))
            @always_comb
            def _pass16():
                sum16.next = sum16_reg[16:]

        if chksum!=None:
            assert len(chksum)>=16, "checksum: expects len(chksum)>={}, but len(chksum)={}".format(16, len(chksum))
            @always_comb
            def _invert():
                chksum.next = ~sum16_reg[16:] & 0xFFFF

    return instances()



if __name__ == '__main__':
    pass




import unittest
from itertools import cycle, dropwhile, islice
from struct import unpack

from myhdl_lib.simulation.payload_generator import payload_generator

class TestPayloadGenerator(unittest.TestCase):


    def testLevels(self):
        ''' PAYLOAD_GENERATOR: Levels '''
        MAX_INT = 67
        MAX_PKT_LEN = 7
        MAX_DIM_SIZE = 14

        for _ in range(10):
            for l in range(0, 3):
                pld = payload_generator(levels=l, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=False, sequential=False)

                pld = [pld] # level -1
                for _ in range(l):
                    # For each level
                    for ls in pld:
                        # check sublevels
                        assert isinstance(ls, list)
                        assert 1<=len(ls)
                        assert len(ls)<=MAX_DIM_SIZE
                    # flatten
                    pld = [x for ls in pld for x in ls]

                # For level 1
                for ls in pld:
                    # check payload length
                    assert isinstance(ls, list)
                    assert 1<=len(ls)
                    assert len(ls)<=MAX_PKT_LEN, "{}".format(ls)
                # flatten
                pld = [x for ls in pld for x in ls]

                # for level 0
                for x in pld:
                    # check payload elements
                    assert isinstance(x, int)
                    assert 0<=x
                    assert x<=MAX_INT


    def testSequential(self):
        ''' PAYLOAD_GENERATOR: Sequential '''
        MAX_INT = 67
        MAX_PKT_LEN = 7
        MAX_DIM_SIZE = 14

        def expected_sequence(l):
            cycled = cycle(range(MAX_INT+1))
            skipped = dropwhile(lambda x: x != 1, cycled)
            sliced = islice(skipped, None, len(pld))
            return list(sliced)

        for i in range(10):
            sequential = (i%2)==0
            # Single packet
            pld = payload_generator(levels=0, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=False, sequential=sequential)
            assert isinstance(pld, list)
            if sequential:
                assert pld==expected_sequence(len(pld))
            else:
                assert pld!=expected_sequence(len(pld))

            # Two packets
            pld = payload_generator(levels=1, dimensions=2, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=False, sequential=sequential)
            assert isinstance(pld, list)
            assert len(pld)==2
            assert isinstance(pld[0], list)
            assert isinstance(pld[1], list)
            # flatten
            pld = [x for ls in pld for x in ls]
            if sequential:
                assert pld==expected_sequence(len(pld))
            else:
                assert pld!=expected_sequence(len(pld))


    def testString(self):
        ''' PAYLOAD_GENERATOR: String '''
        MAX_INT = 67
        MAX_PKT_LEN = 7
        MAX_DIM_SIZE = 14

        for i in range(10):
            string = (i%2)==0
            # Single packet
            pld = payload_generator(levels=0, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=string, sequential=False)
            if string:
                assert isinstance(pld, str)
                for x in pld:
                    assert unpack('B', x)[0] >= 0
                    assert unpack('B', x)[0] <= MAX_INT
            else:
                assert isinstance(pld, list)
                for x in pld:
                    assert isinstance(x, int)
                    assert x >= 0
                    assert x <= MAX_INT

            # Two packets
            pld = payload_generator(levels=1, dimensions=2, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=string, sequential=False)
            assert isinstance(pld, list)
            assert len(pld)==2
            if string:
                assert isinstance(pld[0], str)
                assert isinstance(pld[1], str)
                pld = pld[0]+pld[1]
                for x in pld:
                    assert unpack('B', x)[0] >= 0
                    assert unpack('B', x)[0] <= MAX_INT
            else:
                assert isinstance(pld[0], list)
                assert isinstance(pld[1], list)
                pld = pld[0]+pld[1]
                for x in pld:
                    assert isinstance(x, int)
                    assert x >= 0
                    assert x <= MAX_INT


    def testDimensions(self):
        ''' PAYLOAD_GENERATOR: Dimensions '''
        MAX_INT = 67
        MAX_PKT_LEN = 7
        MAX_DIM_SIZE = 14

        lvls = 0
        dim = None
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, str)

        dim = 10
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, str)
        assert len(pld)==dim

        dim = [10]
        with self.assertRaises(TypeError):
            payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)

        lvls = 1
        dim = None
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        for ls in pld:
            assert isinstance(ls, str)

        dim = 3
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==dim
        for ls in pld:
            assert isinstance(ls, str)

        dim = [4,6,9]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls, d in zip(pld,dim):
            assert isinstance(ls, str)
            assert len(ls)==d

        dim = [4,None,9]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls in pld:
            assert isinstance(ls, str)

        dim = [4,6,[9]]
        with self.assertRaises(TypeError):
            payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)

        lvls = 2
        dim = None
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        for ls in pld:
            assert isinstance(ls, list)
            for s in ls:
                assert isinstance(s, str)

        dim = 5
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==dim
        for ls in pld:
            assert isinstance(ls, list)
            for s in ls:
                assert isinstance(s, str)

        dim = [2,3,4]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls in pld:
            assert isinstance(ls, list)
            for s in ls:
                assert isinstance(s, str)

        dim = [[2],[3,4],[5,6,7]]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls, d in zip(pld, dim):
            assert isinstance(ls, list)
            assert len(ls)==len(d)
            for s, dd in zip(ls,d):
                assert isinstance(s, str)
                assert len(s)==dd

        dim = [[2],[3,None],[5,6,7]]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls, d in zip(pld, dim):
            assert isinstance(ls, list)
            assert len(ls)==len(d)
            for s in ls:
                assert isinstance(s, str)

        dim = [[2],None,[5,6,7]]
        pld = payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)
        assert isinstance(pld, list)
        assert len(pld)==len(dim)
        for ls in pld:
            assert isinstance(ls, list)
            for s in ls:
                assert isinstance(s, str)

        dim = [[2],[3,[4]],[5,6,7]]
        with self.assertRaises(TypeError):
            payload_generator(levels=lvls, dimensions=dim, max_int=MAX_INT, max_pkt_len=MAX_PKT_LEN, max_dim_size=MAX_DIM_SIZE, string=True, sequential=False)


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


import unittest

from myhdl import *
from myhdl_lib.mux import mux
import myhdl_lib.simulation as sim

import random




class TestMux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]


    def testMux5(self):
        """ MUX: 5 inputs"""
        DMAX = 100
        NUM_INPUTS = 5

        def mux_top(sel, di_0, di_1, di_2, di_3, di_4, do):
            ''' Needed when mux is co-simulated as top level'''
            ls_di = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_INPUTS)]
            @always_comb
            def _assign():
                ls_di[0].next = di_0
                ls_di[1].next = di_1
                ls_di[2].next = di_2
                ls_di[3].next = di_3
                ls_di[4].next = di_4

            inst = mux(sel=sel, ls_di=ls_di, do=do)
            return instances()

        sel = Signal(intbv(0,min=0,max=NUM_INPUTS))
        ls_di = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_INPUTS)]
        do = Signal(intbv(0,min=0,max=DMAX))

        argl = {"sel":sel, "di_0":ls_di[0], "di_1":ls_di[1], "di_2":ls_di[2], "di_3":ls_di[3], "di_4":ls_di[4], "do":do}

        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_INPUTS):
                    ls_di[i].next = 0
                yield delay(10)

                for _ in range(10):
                    d = [random.randint(0,DMAX-1) for _ in range(NUM_INPUTS)]
                    for i in range(NUM_INPUTS):
                        ls_di[i].next = d[i]

                    for s in range(NUM_INPUTS):
                        sel.next = s
                        yield delay(10)
                        assert d[s] == do, "Mux output (sel={}): expected {}, detected {}".format(sel, d[s], do)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(mux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm


    def testMux2(self):
        """ MUX: 2 inputs, boolean Select"""
        DMAX = 100
        NUM_INPUTS = 2

        def mux_top(sel, di_0, di_1, do):
            ''' Needed when mux is co-simulated as top level'''
            ls_di = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_INPUTS)]
            @always_comb
            def _assign():
                ls_di[0].next = di_0
                ls_di[1].next = di_1

            inst = mux(sel=sel, ls_di=ls_di, do=do)
            return instances()

        sel = Signal(bool(0))
        ls_di = [Signal(intbv(0,min=0,max=DMAX)) for _ in range(NUM_INPUTS)]
        do = Signal(intbv(0,min=0,max=DMAX))

        argl = {"sel":sel, "di_0":ls_di[0], "di_1":ls_di[1], "do":do}

        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_INPUTS):
                    ls_di[i].next = 0
                yield delay(10)

                for _ in range(10):
                    d = [random.randint(0,DMAX-1) for _ in range(NUM_INPUTS)]
                    for i in range(NUM_INPUTS):
                        ls_di[i].next = d[i]

                    for s in range(NUM_INPUTS):
                        sel.next = s
                        yield delay(10)
                        assert d[s] == do, "Mux output (sel={}): expected {}, detected {}".format(sel, d[s], do)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(mux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm



if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testMux']
    unittest.main()



from myhdl import *


def assign(a,b):
    ''' Combinatorial assignment: a = b '''
    @always_comb
    def _assign():
        a.next = b
    return _assign


def byteorder(bv_di, bv_do, REVERSE=True):
    """ Reverses the byte order of an input bit-vector
            bv_di - (i) input bit-vector
            bv_do - (o) output bit-vector
    """
    BITS  = len(bv_di)

    assert (BITS % 8)==0, "byteorder: expects len(bv_in)=8*x, but len(bv_in)={}".format(BITS)

    if REVERSE:
        BYTES = BITS//8
        y = ConcatSignal(*[bv_di(8*(b+1),8*b) for b in range(BYTES)])

        @always_comb
        def _reverse():
            bv_do.next = y

    else:
        @always_comb
        def _pass():
            bv_do.next = bv_di

    return instances()


if __name__ == '__main__':
    pass


import unittest

from myhdl import *
from myhdl_lib.handshake import hs_demux
import myhdl_lib.simulation as sim


class TestHSDemux(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.simulators = ["myhdl", "icarus"]

    def testHSDemux3(self):
        ''' HS_DEMUX: 3 outputs '''
        NUM_OUTPUTS = 3

        def hs_demux_top(sel, i_rdy, i_vld, o0_rdy, o0_vld, o1_rdy, o1_vld, o2_rdy, o2_vld):
            ''' Needed when hs_demux is co-simulated as top level'''

            hsi_rdy = Signal(bool(0))
            hsi_vld = Signal(bool(0))
            hso0_rdy = Signal(bool(0))
            hso0_vld = Signal(bool(0))
            hso1_rdy = Signal(bool(0))
            hso1_vld = Signal(bool(0))
            hso2_rdy = Signal(bool(0))
            hso2_vld = Signal(bool(0))

            @always_comb
            def _assign():
                hso0_rdy.next = o0_rdy
                hso1_rdy.next = o1_rdy
                hso2_rdy.next = o2_rdy

                o0_vld.next = hso0_vld
                o1_vld.next = hso1_vld
                o2_vld.next = hso2_vld

                i_rdy.next = hsi_rdy
                hsi_vld.next = i_vld

            hsi = (hsi_rdy, hsi_vld)
            ls_hso = [(hso0_rdy, hso0_vld), (hso1_rdy, hso1_vld), (hso2_rdy, hso2_vld)]

            _inst = hs_demux(sel, hsi=hsi, ls_hso=ls_hso)

            return instances()

        hsi_rdy = Signal(bool(0))
        hsi_vld = Signal(bool(0))
        hsi = (hsi_rdy, hsi_vld)

        ls_hso_rdy = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso_vld = [Signal(bool(0)) for _ in range(NUM_OUTPUTS)]
        ls_hso = zip(ls_hso_rdy, ls_hso_vld)

        sel = Signal(intbv(0)[NUM_OUTPUTS:])

        argl = {"sel":sel,
                "i_rdy":hsi[0], "i_vld":hsi[1],
                "o0_rdy":ls_hso[0][0], "o0_vld":ls_hso[0][1],
                "o1_rdy":ls_hso[1][0], "o1_vld":ls_hso[1][1],
                "o2_rdy":ls_hso[2][0], "o2_vld":ls_hso[2][1]}

        def stim():
            @instance
            def _inst():
                sel.next = 0
                for i in range(NUM_OUTPUTS):
                    ls_hso_rdy[i].next = 0
                hsi_vld.next = 0
                yield delay(10)
                for i in range(NUM_OUTPUTS):
                    assert 0==ls_hso_vld[i], "hso_vld[{}]: expected {}, detected {}".format(i, 0, ls_hso_vld[i])
                assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format( 0, hsi_rdy)

                for s in range(NUM_OUTPUTS):
                    sel.next = s
                    x = intbv(0)[NUM_OUTPUTS+1:]
                    for i in range(x.max):
                        x[:] = i
                        for k in range(NUM_OUTPUTS):
                            ls_hso_rdy[k].next = x[k]
                        hsi_vld.next = x[NUM_OUTPUTS]

                        yield delay(10)


                        for k in range(NUM_OUTPUTS):
                            if (s==k) and (x[NUM_OUTPUTS]==1):
                                assert 1==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 1, ls_hso_vld[k])
                            else:
                                assert 0==ls_hso_vld[k], "hso_vld[{}]: expected {}, detected {}".format(k, 0, ls_hso_vld[k])

                        if x[s]==1:
                            assert 1==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(1, hsi_rdy)
                        else:
                            assert 0==hsi_rdy, "hsi_rdy: expected {}, detected {}".format(0, hsi_rdy)

                yield delay(10)
                raise StopSimulation
            return _inst

        getDut = sim.DUTer()

        for s in self.simulators:
            getDut.selectSimulator(s)

            dut=getDut(hs_demux_top, **argl)
            stm = stim()
            Simulation( dut, stm).run()
            del dut, stm


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()